# 文件系统

> 注意：此文档的重心不在于介绍FAT32文件系统的基本概念

## 基本

在设计文件系统之初，我们已经考虑过是否有引入VFS的需求。从编程者角度出发，引入VFS的机制将会大大地方便上层接口的编程，但是在底层需要增加些许的兼容性代码。但是由于时间精力有限，我们打算先将FAT32文件系统实现，在接口设计上保留VFS的风格与兼容性，方便后续的「移植」。

我们较为完善地实现FAT32文件系统，包括：
* 全面支持长短文件名目录项读取与写入（包含校验和）
* 文件的创建删除以及读写截断
* 簇的申请与释放

具体的，针对文件名大小写的问题，我们学习Linux上的存储方式，通过长短目录项配合的方法来实现我们的目的。

在实现过程中，我们发现了许多遍历簇的过程存在重复，具有可复用的潜力，但是实际复用过程并非一帆风顺：尽管大体过程类似，但是其中存在很多细微的区别。因此我们的复用目标应该是在尽可能保持程序最大化复用的前提下，还要保留功能接口对细节的控制能力。那么我们该如何做呢？答案很简单，通过函数指针的形式（Talk is cheap, show me the code:
```c
FR_t fat_travs_dir(fat32_t *fat, uint32_t dir_clus, uint32_t dir_offset,
                          entry_handler_t handler, 
                          int alloc, uint32_t *offset,
                          void *p1, void *p2) 
{
    // ...
    while(1) { // 遍历簇
        if(dir_offset > BPC(fat)) {
            dir_offset -= BPC(fat);
            goto next_clus;
        }
        // 簇起始扇区
        int i, j;
        uint32_t sec = clus2datsec(fat, curr_clus);
        for(i = dir_offset / BPS(fat), dir_offset %= BPS(fat);  // 遍历簇内扇区
                i < fat->sec_per_cluster;   i++  ) 
        { 
            buf_t *b = bread(fat->dev, sec + i);
            for(j = dir_offset / item_size, dir_offset = 0;     // 遍历扇区内目录项
                j < item_per_sec;   j++  ) 
            { 
                FR_t res;
                uint32_t ofst = (clus_cnt * SPC(fat) + i) * BPS(fat) + j * item_size;
                dir_item_t *item = (dir_item_t *)(b->data) + j;
                travs_meta_t meta = {.buf = b, .item_no = j, .nitem = item_per_sec,.offset = ofst};
                if((res = handler(item, &meta, p1, p2)) != FR_CONTINUE) {
                    brelse(b);
                    if(offset) // 计算当前目录项在当前目录数据簇中的偏移量
                        *offset = ofst;
                    return res; 
                }
            }
            brelse(b);
        }
    next_clus:
        // ...
    }

    return FR_ERR;
}
```
代码的中心位置是我们的主角handler函数指针，外部框架代码来完成琐碎的遍历目录工作，然后将核心信息与元数据包装后传递给具体的handler函数，内部函数来通过参数与**返回值**的方式来实现数据获取与流程控制:
```c
// 返回值
typedef enum _FAT_RESULT_t{
	FR_OK,
	FR_ERR,
	FR_CONTINUE, // 用于目录继续遍历的标识
} FR_t;
```

依靠强大的框架代码，我们仅仅用了900行代码就实现了整个完整的FAT32文件系统功能！基于初始的设计目标，我们的FAT32文件系统实现了对外的完全解耦，甚至可以独立地将此文件抽离来供外部系统使用！

## 缓存设计

文件系统在底层仍然保留了xv6的块缓存机制（buffer cache），此外，我们还设计了页缓存机制。

